/*****************************
* Source : xmlparser.prg
* System : <unkown>
* Author : Phil Ide
* Created: 12/02/2004
*
* Purpose:
* ----------------------------
* History:
* ----------------------------
*    12/02/2004 13:14 PPI - Created
*****************************/

#include "Xbp.ch"
#include "error.ch"
#include "Common.ch"

#define PIXML_SCAN_SHALLOW   0
#define PIXML_SCAN_DEEPSCAN  1

#define CRLF Chr(13)+Chr(10)
#define WS_INTAG     ' '+Chr(9)+Chr(0)+CRLF
#define WS_OUTTAG    WS_INTAG
#define WS_CONTENT   ''

#define OUT_TAG      0
#define IN_TAG       1
#define IN_CONTENT   2

// TAG_PAD provides indentation on output
#define TAG_PAD 0 //4

#ifdef _TEST_
   Function Main(cFile)
      local oXml
      local nH
      if FExists(cFile)
         oXml := ParseXml( memoread(cFile) )
         if !oXml == NIL
            nH := FCreate('test.xml')
            FWrite(nH,oXml:asString())
            FClose(nH)
         endif
      endif
      return nil
#endif


Function ParseXml( cStream )
   local oParser := XmlParser():new(cStream)
   local oXml

   oXml := oParser:parse()
   return oXml

CLASS XMLParser
   EXPORTED:
      VAR cData
      VAR iPos
      VAR inTag
      VAR nEnd
      VAR inString
      VAR stringDelim
      VAR LineNum
      VAR oTree
      VAR lastPos

      METHOD init
      METHOD parse

   PROTECTED:
      METHOD XMLParse
      METHOD getNextToken
      METHOD lookAhead
      METHOD strip
      METHOD ungetToken
      METHOD xmlError
      METHOD getTag
ENDCLASS


METHOD XMLParser:init(cStream)
   ::cData       := cStream
   ::iPos        := 1
   ::inTag       := OUT_TAG
   ::nEnd        := Len(::cData)
   ::inString    := FALSE
   ::stringDelim := {}
   ::LineNum     := 1
   ::oTree       := {}
   return self

METHOD XMLParser:parse()
   local oDoc
   // call lexical analyser
   ::XmlParse()
   if !Empty(::oTree)
      oDoc := ::oTree[1]
   endif
   return oDoc

METHOD XMLParser:XmlParse()
   local cToken := '0'
   local oDoc
   local n

   While !Empty( cToken := ::getNextToken() )
      if cToken == '<'
         ::getTag() // call lexical parser

      elseif ::inTag == OUT_TAG .and. !Empty(::oTree)
         // reset pointer
         ::iPos = ::lastPos
         n := ::lookAhead()
         cToken := SubStr( ::cData, ::iPos, n )
         if !Empty(::Strip(cToken,WS_OUTTAG)) // is there only white-space?
            ATail(::oTree):content := cToken  // keep white-space if there is data
         endif
         ::iPos += n
      else
         ::xmlError(cToken)
      endif
   Enddo
   return nil

METHOD XMLParser:getNextToken()
   local cToken := ''
   local cA

   ::lastPos := ::iPos

   While ::iPos <= ::nEnd
      cA := ::cData[::iPos]
      if ::cData[::iPos] == Chr(10)
         ::LineNum++
      endif

      do case

         case ::inString .and. !(::cData[::iPos] == ATail(::stringDelim))
            cToken += ::cData[::iPos]

         case ::cData[::iPos] == '?' .and. ::inTag == IN_TAG .and. !::inString
            if Len(cToken) > 0
               exit
            else
               cToken := '?'
               ::iPos++
               exit
            endif

         case ::cData[::iPos] == '=' .and. ::inTag == IN_TAG
            if Len(cToken) > 0
               exit
            else
               cToken := '='
               ::iPos++
               exit
            endif

         case ::inTag == OUT_TAG .and. ::cData[::iPos] == '<' .and. !::inString
            if Len(cToken) > 0
               exit
            else
               cToken := '<'
               ::inTag := IN_TAG
               ::iPos++
               exit
            endif

         case ::inTag == OUT_TAG .and. !(::cData[::iPos] $ '"'+"'") .and. !(::cData[::iPos] $ WS_OUTTAG)
            cToken += ::cData[::iPos]

         case ::cData[::iPos] == '>' .and. !::inString
            if Len(cToken) > 0
               exit
            else
               cToken := '>'
               ::inTag := OUT_TAG
               ::iPos++
               exit
            endif

         case ::cData[::iPos] == '"' .or. ::cData[::iPos] == "'"
            if !::inString
               aadd(::stringDelim, ::cData[::iPos])
               ::inString := TRUE
            elseif ::cData[::iPos] == ATail(::stringDelim)
               ASize(::stringDelim, Len(::stringDelim)-1)
               ::inString := FALSE
            endif
            cToken += ::cData[::iPos]

         case ::inTag == IN_TAG .and. ::cData[::iPos] $ WS_INTAG
            if Len(cToken) > 0
               ::iPos++
               exit
            endif

         case ::inTag == OUT_TAG .and. ::cData[::iPos] $ WS_OUTTAG
            if Len(cToken) > 0
               ::iPos++
               exit
            endif

         otherwise
            cToken += ::cData[::iPos]
      endcase
      ::iPos++
   Enddo
   return cToken

METHOD XMLParser:lookAhead()
   local i := ::iPos+1
   local n := 1
   local aDelim := {}

   While i <= ::nEnd
      if ::cData[i] $ '"'+"'"
         if !Empty(aDelim) .and. ::cData[i] == ATail(aDelim)
            ASize(aDelim, Len(aDelim)-1)
         else
            aadd(aDelim, ::cData[i])
         endif
      elseif ::cData[i] $ '<'
         exit
      endif
      i++
   Enddo
   return i-::iPos

METHOD XMLParser:Strip(cToken,cStrip)
   local i
   local n := Len(cToken)
   local c := ''

   for i := 1 to n
      if !(cToken[i] $ cStrip)
         c += cToken[i]
      endif
   next
   return c


METHOD XMLParser:ungetToken()
   ::iPos := ::lastPos
   return NIL



METHOD XMLParser:XmlError(cToken)
   local o := Error():new()

   o:args := {cToken, ::LineNum, ::iPos-1}
   o:description := "XML Parsing error - Error in Input"
   o:severity := XPP_ES_FATAL
   o:subSystem := "XML"
   Eval( ErrorBlock(), o)
   return nil

METHOD XMLParser:getTag()
   local oTag := Xml():new()
   local cAttrName
   local cAttrVal
   local cName
   local cToken

   oTag:name := ::getNextToken()
   if oTag:name == '?' .or. oTag:name == '!' .or. oTag:name == '/'
      oTag:name += ::getNextToken() // ?xml
   endif

   While !(::cData[::iPos] == '>')
      cAttrName := ::getNextToken()
      cAttrVal := NIL
      if !(cAttrName == '?' .or. cAttrName == '/')
         cToken := ::getNextToken() // remove '='
         if cToken == '='
            cAttrVal := ::getNextToken()
            // strip out quotes
            if cAttrVal[1] $'"'+"'"
               cAttrVal := SubStr(cAttrVal,2)
            endif
            if cAttrVal[-1] $'"'+"'"
               cAttrVal := Left(cAttrVal,Len(cAttrVal)-1)
            endif
         else
            ::ungetToken(cToken)
         endif
         aadd( oTag:attribute, {cAttrName, cAttrVal} )
      elseif cAttrName == '/'
         oTag:hasEndTag := FALSE
      else
         aadd( oTag:attribute, {cAttrName, NIL} )
      endif
   enddo
   if !(oTag:name[1] == '/')
      if !(cAttrName == '?' .or. cAttrName == '/')
         if !Empty(::oTree)
            oTag:parent := ATail(::oTree)
            aadd( ATail(::oTree):children, oTag )
         endif
         aadd(::oTree, oTag)
      elseif cAttrName == '?' .and. Empty(::oTree)
         oTag:hasEndTag := FALSE
         aadd(::oTree,oTag)
      elseif lower(oTag:name) == '?xml'
         aadd(::oTree,oTag)
      elseif oTag:name[1] == '!'
         oTag:hasEndTag := FALSE
      elseif cAttrName[1] == '/'
         oTag:hasEndTag := FALSE
         oTag:parent := ATail(::oTree)
         aadd( ATail(::oTree):children, oTag )
      endif
   else
      oTag:hasEndTag := FALSE
      ASize(::oTree, Len(::oTree)-1)
   endif

   ::getNextToken() // clear '>'
   return oTag

CLASS XML
   EXPORTED:
      VAR parent
      VAR name
      VAR attribute
      VAR content
      VAR children
      VAR hasEndTag
      VAR childID

      METHOD init
      METHOD asString

      METHOD compose IS asString
      METHOD parent
      METHOD children
      METHOD getChild
      METHOD allSiblings
      METHOD siblings
      METHOD getSibling
      METHOD getPreviousSibling
      METHOD getNextSibling
      METHOD isOrphan
      METHOD getAttribute
      METHOD setAttribute
      METHOD findAttribute
      METHOD findChildFromName
      METHOD findChildFromAttribute
      METHOD findAllChildrenFromName
      METHOD findSiblingFromName
      METHOD findSiblingFromAttribute


ENDCLASS

METHOD Xml:init()
   ::name      := ''
   ::attribute := {}
   ::children  := {}
   ::hasEndTag := TRUE
   return self

METHOD Xml:parent()
   return ::parent

METHOD Xml:children()
   return ::children

METHOD Xml:getChild(n)
   local oRet

   if valType(n) == 'N' .and. n > 0 .and. n <= len(::Children)
      oRet := ::Children[n]
   endif
   return oRet   

METHOD Xml:allSiblings()
   return ::parent:children

METHOD Xml:siblings()
   local oParent := ::parent
   local i
   local aRet := {}

   for i := 1 to len(oParent:children)
      if !(oParent:children[i] == self)
         aadd( aRet, oParent:children[i] )
      endif
   next
   return aRet

METHOD Xml:getSibling(n)
   local oRet

   if ValType(n) == 'N' .and. n > 0 .and. n <= len(::parent:children)
      oRet := ::parent:children[n]
   endif
   return oRet

METHOD Xml:getPreviousSibling(nId)
   local i
   local oRet

   default nId to ::id

   i := AScan(::parent:children, {|o| o == self })
   if i > 1
      oRet := ::parent:children[i-1]
   endif
   return (oRet)

METHOD Xml:getNextSibling(nId)
   local i
   local oRet

   default nId to ::id

   i := AScan(::parent:children, {|o| o == self })
   if i < len(::parent:children)
      oRet := ::parent:children[i+1]
   endif
   return (oRet)

METHOD Xml:isOrphan()
   return ::parent == NIL


METHOD Xml:getAttribute(x)
   local cRet

   cRet := ::attribute[x]
   return cRet

METHOD Xml:setAttribute(x, xValue)
   local aRet
   local i

   do case
      case ValType(x) == 'N'
         if x > 0 .and. x <= len(::attribute)
            ::attribute[x][2] := Var2Char(xValue)
            aRet := ::attribute[x]
         endif

      case ValType(x) == 'C'
         i := Ascan( ::attribute, {|e| lower(e[1]) == lower(x) } )
         if i == 0
            aadd(::attribute,{x,Var2Char(xValue)})
            i := len(::attribute)
         else
            ::attribute[i][2] := Var2Char(xValue)
         endif
         aRet := ::attribute[i]
   endcase
   return aRet

METHOD Xml:findAttribute(c)
   local i := Ascan( ::attribute, {|e| lower(e[1]) == lower(c) } )
   local cRet

   if i > 0
      cRet := ::attribute[i][2]
   endif
   return cRet

METHOD Xml:findChildFromName( cName, nScanLevel )
   local nDepth := 0
   local oRet
   local i

   for i := 1 to len(::children)
      if lower(::children[i]:name) == lower(cName)
         oRet := ::children[i]
      elseif nScanLevel == PIXML_SCAN_DEEPSCAN
         oRet := ::children[i]:findChildFromName( cName, nScanLevel )
      endif
      if !(oRet == NIL)
         exit
      endif
   next
   return oRet

METHOD Xml:findChildFromAttribute( cAttr, nScanLevel )
   local nDepth := 0
   local oRet
   local aRet
   local i

   for i := 1 to len(::children)
      aRet := ::children[i]:findAttribute(cAttr)
      if aRet == NIL .and. nScanLevel == PIXML_SCAN_DEEPSCAN
         oRet := ::children[i]:findChildFromAttribute( cAttr, nScanLevel )
      else
         oRet := ::children[i]
      endif
      if !(oRet == NIL)
         exit
      endif
   next
   return oRet

METHOD Xml:findSiblingFromName(cName)
   local aSiblings := ::siblings()
   local i
   local oRet

   for i := 1 to len(aSiblings)
      if lower(aSiblings[i]:name) == lower(cName)
         oRet := aSiblings[i]
         exit
      endif
   next
   return oRet

METHOD Xml:findSiblingFromAttribute(cAttr)
   local aSiblings := ::siblings()
   local i
   local c
   local oRet

   for i := 1 to len(aSiblings)
      c := aSiblings[i]:findAttribute(cAttr)
      if !(c == NIL)
         oRet := aSiblings[i]
         exit
      endif
   next
   return oRet

METHOD Xml:findAllChildrenFromName(cName)
   local aRet := {}
   local i
   local n
   local aTmp

   for i := 1 to len(::children)
      if lower(::children[i]:name) == lower(cName)
         aadd( aRet, ::children[i] )
      else
         aTmp := ::children[i]:findAllChildrenFromName(cName)
         for n := 1 to len(aTmp)
            aadd( aRet, aTmp[n] )
         next
      endif
   next
   return aRet
   

METHOD Xml:asString(nDepth)
   local cRet
   local i
   local v
   local n

   default nDepth to 0

   cRet := Space(TAG_PAD*nDepth)+'<'

   cRet += ::name

   v := Len(::attribute)
   for i := 1 to v
      cRet += ' '+::attribute[i][1]
      if !(::attribute[i][2] == NIL)
         cRet += '="'+::attribute[i][2]+'"'
      endif
   next
   if ::hasEndTag .or. ::name[1] == '?'
      cRet += '>'
   endif

   if !Empty(::content)
      cRet += ::content
   endif
   if !Empty(::children)
      cRet += CRLF
   endif

   for i := 1 to Len(::children)
      n := iif( lower(::name) == '?xml', 0, nDepth+1 )
      cRet += ::children[i]:asString(n)
   next
   if !(::name[1] $ '?!')
      if ::hasEndTag
         if !Empty(::children)
            cRet += Space(TAG_PAD*nDepth)+'</'
         else
            cRet += '</'
         endif

         cRet += ::name+'>'+CRLF
      else
         cRet += ' />'+CRLF
      endif
   endif
   return cRet



